"""
Early draft of a very handy utility that takes a CCDL file (Concise Campaign Definition Language)
and creates a graph(viz) visualization of it.

"""
import sys
#import pygraphviz # This makes emod-api dependent on graphviz. Might be nice if this was optional?
#import graphviz # This makes emod-api dependent on graphviz. Might be nice if this was optional?
import networkx as graphviz
from emod_api.interventions.ccdl.ccdl import *
from math import gcd

debug = True


def get_nickname_from_event( event_num, pieces ):
    """
    Allow nodes to get briefer and potentially more helpful nicknames. Default will probably remain a
    the nasty autogen above. Users can override this function with a callback of their own.
    """
    # If user has manually added an optional 5th 'column', use that, but turns certain characters into newlines.
    if len(pieces)==5:
        event_name = pieces[4].strip().replace( " ", "\n" ).replace("->","%%%").replace( "-", "\n" ).replace( "%%%", "->" )
    else:
        #event_name = f"[{event_num}]{pieces[WHEN_IDX]}*{pieces[WHERE_IDX]}*{pieces[WHO_IDX]}"
        main_iv = pieces[WHAT_IDX].split("+")[0].split("->")[-1]
        main_iv = main_iv.replace( "(", "\n(" )
        event_name = f"[{event_num}]{main_iv}@{pieces[WHEN_IDX]}@{pieces[WHO_IDX]}"
        event_name = event_name.replace( "@", "@\n" )
    return event_name 


def get_colour_from_event( tokens ): 
    """
    Allow nodes to get a content-dependent colour. Default to just white. Users can override this function with 
    a callback of their own. Have been using colour to capture IP categories. 
    """
    return "white" 

def get_shape_from_event( tokens ):
    """
    Allow nodes to get a content-dependent shape. Default to circle. Users can override this function with 
    a callback of their own. Have been using shape to capture 'epoch' categories. Possible shapes include ellipse, circle, square, and diamond.  Full list can be found at: https://www.graphviz.org/doc/info/shapes.html
    """
    #return "circle" 
    return "ellipse" 


def set_beautifiers( name_cb=None, colour_cb=None, shape_cb=None ):
    """
    Override default no-op callbacks for setting nicknames, colours, and shapes of campaign nodes
    """
    global get_nickname_from_event, get_colour_from_event, get_shape_from_event
    if name_cb:
        get_nickname_from_event = name_cb
    if colour_cb:
        get_colour_from_event = colour_cb
    if shape_cb:
        get_shape_from_event = shape_cb
    return

def viz( in_name = "campaign.ccdl", out_name = "camp.sv", display = True, whitelist = None ): 
    ccdl = open( in_name ).readlines()
    #dot = graphviz.Digraph(out_name, comment='Patient Pathway Design', format="svg")
    dot = graphviz.DiGraph(format="svg")
    #dot.graph_attr.update(size="10,10")

    producers = {}
    consumers = {}
    # Don't see how to avoid a specific list here. Not complete, needs more diagnostics.
    broadcasters = ["BroadcastEvent", "SimpleHealthSeekingBehavior", "DiagnosticTreatNeg", "SimpleDiagnostic", "HIVPiecewiseByYearAndSexDiagnostic", "HIVRandomChoice", "HIVMuxer", "STIIsPostDebut", "HIVRapidHIVDiagnostic", "HIVARTStagingCD4AgnosticDiagnostic", "HIVSigmoidByYearAndSexDiagnostic", "HIVDelayedIntervention" ]

    event_names = {}
    event_num = 0
    for camp_event in ccdl:
        if debug:        
            print( f"Processing line {camp_event}." )

        # We're going to do a brute force purge of DelayedInterventions for now.
        pieces = camp_event.strip().split( main_sep )
        if len(pieces)<2:
            event_num += 1
            continue

        if "Outbreak" in pieces[-1]:
            # It is useful to 'hack' a few lines so that major infection lifecycle events get tied into
            # our visualization.
            camp_event = camp_event.replace( pieces[-1], pieces[-1]+"+BroadcastEvent(TBActivation)+BroadcastEvent(NewInfection)+BroadcastEvent(NewInfectionEvent)+BroadcastEvent(Symptomatic)" )
            pieces = camp_event.strip().split( main_sep )

        # Color represents IP
        # Shape represents Epoch
        node_name = get_nickname_from_event( event_num, pieces )
        event_names[event_num] = node_name
        node_color = get_colour_from_event(pieces)
        node_shape = get_shape_from_event(pieces)

        if debug:
            print( f"Creating node with name {event_names[event_num]}." )
        #dot.node( name=event_names[event_num], style='filled', fillcolor = node_color, shape=node_shape )
        triggered = pieces[WHAT_IDX].split( post_trigger_sep )
        if len(triggered)>1:
            triggers = triggered[0].strip()
            for trigger in triggers.split( multi_trigger_sep ):
                if trigger not in consumers:
                    consumers[trigger] = set() 
                consumers[trigger].add( event_num )

        triggeree = triggered[-1].split( post_delay_sep )[-1].strip()
        for broadcaster in broadcasters:
            sender = (triggeree.split( multi_iv_sep )[-1]).strip()
            if broadcaster in sender:
                print( f"'sender' {broadcaster} found in 'what?'" )
                from parse import parse
                regex = broadcaster + "({})"
                print( f"Looking for {regex} in {sender}" )
                tokens = parse( regex, sender )
                if tokens:
                    signals = tokens[0]
                    for signal in signals.split( multi_signal_sep ):
                        if debug:
                            print( f"Found an event broadcasting {signal}" )
                        if signal not in producers:
                            producers[signal] = set()
                        producers[signal].add( event_num )
                else:
                    print( f"ERROR: Oh darn. Searching for {regex} in {sender} produced None." )
        event_num += 1

    if debug:
        print( "PRODUCERS\n" )
        print( producers )
        print( "\nCONSUMERS\n" )
        print( consumers )

    connecteds = set()
    connections = []
    # This is a list so we can do more than 1 eventually but for now lets just do 1.
    event_whitelist = []
    if whitelist:
        event_whitelist.append( whitelist )

    for event in producers:
        if event_whitelist and event not in event_whitelist:
            continue
        senders = producers[event]
        if event in consumers:
            receivers = consumers[event]
            for from_event in senders:
                for to_event in receivers:
                    # Reject edges where IPs don't "overlap"
                    who_from = ccdl[from_event].split( main_sep )[2]
                    sex_from = "Both"
                    ip_from = who_from.split("/")[-1] if "/" in who_from else "*"
                    ip_from = ip_from.replace( ":", "=" )
                    if ip_from == "Male" or ip_from == "Female":
                        sex_from = ip_from
                    if ">" in ip_from or "<" in ip_from or "ale" in ip_from:
                        ip_from = "*"

                    who_to = ccdl[to_event].split( main_sep )[2]
                    sex_to = "Both"
                    ip_to = who_to.split("/")[-1] if "/" in who_to else "*"
                    ip_to = ip_to.replace( ":", "=" ) 
                    if ip_to == "Male" or ip_to == "Female":
                        sex_to = ip_to
                    if ">" in ip_to or "<" in ip_to or "ale" in ip_to:
                        ip_to = "*"

                    if debug:
                        if sex_from and sex_to:
                            print( f"This campaign event includes sex-targeting: from={sex_from}, to={sex_to}." )

                    #if ip_to != "*" and ip_from != "*" and ip_from.split(":")[0] == ip_to.split(":")[0] and ip_from.split(":")[1] != ip_to.split(":")[1]:
                    if ip_to != "*" and ip_from != "*":
                        ip_from_key = ip_from.split("=")[0]
                        ip_from_value = ip_from.split("=")[1]
                        ip_to_key = ip_to.split("=")[0]
                        ip_to_value = ip_to.split("=")[1]
                        if ip_from_key == ip_to_key and ip_from_value != ip_to_value: # and sex_from != sex_to: sex comp seemed right but no effect so still unsure
                            if debug:
                                print( f"Mismatch: ip_from_key = {ip_from_key}, ip_to_key = {ip_to_key}, ip_from_value = {ip_from_value}, ip_to_value = {ip_to_value}." )
                            continue # aka do not continue

                    # Reject edges where Times don't "overlap"
                    time_from = ccdl[from_event].split( main_sep )[0]
                    if "(" in time_from:
                        time_from = time_from.split( "(" )[0]
                    time_to = ccdl[to_event].split( main_sep )[0]
                    time_from_start = time_from
                    time_to_start = time_to
                    time_from_end = 1e9
                    time_to_end = 1e9
                    if "-" in time_from:
                        time_from_start = float(time_from.split("-")[0])
                        time_from_end = float(time_from.split("-")[1])
                    if "(" in time_from:
                        time_from_start = float(time_from.split("(")[0])
                        #time_from_end = float(time_from.split("(")[1])
                    if "-" in time_to:
                        time_to_start = float(time_to.split("-")[0])
                        time_to_end = float(time_to.split("-")[1])
                    if "(" in time_to:
                        time_to_start = float(time_to.split("(")[0])
                        #time_to_end = float(time_to.split("-")[1])
                    time_from_start = float(time_from_start)
                    time_to_start = float(time_to_start)
                    if time_from_end <= time_to_start or time_from_start >= time_to_end:
                        if debug:
                            print( f"Discarding edge from {from_event} to {to_event} because not overlapping in time." )
                            print( time_from_start, time_from_end, time_to_start, time_to_end )
                        continue
                    if sex_from != sex_to and sex_to != "Both" and sex_from != "Both":
                        if debug:
                            print( f"Discarding edge from {from_event} to {to_event} because not overlapping in sex." )
                            print( time_from_start, time_from_end, time_to_start, time_to_end )
                        continue
                    if debug:
                        #print( f"Creating edge from {from_event} to {to_event} for {event}." )
                        #print( f"{from_event} ---{event} ---> {to_event}" )
                        print( f"{event_names[from_event]} ---{event} ---> {event_names[to_event]}" )
                    #dot.edge( from_event, to_event, name=event )
                    #dot.edge( str(event_names[from_event]), str(event_names[to_event]), label=event )
                    dot.add_edge( str(event_names[from_event]), str(event_names[to_event]), label=event )
                    connecteds.add( event_names[from_event] )
                    connecteds.add( event_names[to_event] )
                    connections.append( dict({ "from": from_event, "to": to_event, "signal": event }) )

    """
    for event_id, node_name in event_names.items():
        if node_name not in connecteds:
            import pdb
            pdb.set_trace()
            dot.delete( node_name )
    """
    # Honestly not sure what the most general-purpose display solution is for graphviz
    # This seems to generate a pdf and open the registered pdf viewer if there is one.
    #dot.render( view=True )
    #dot.render( out_name )
    #dot.render( directory="/var/opt/idm/apache", filename="camp.sv.png" )

    clusters = list(graphviz.weakly_connected_components(dot)) 
    print( f"Found {len(clusters)} clusters." )

    import pathlib
    import matplotlib.pyplot as plt
    import graphviz as gv
    counter = 0
    
    for cluster in clusters:
        out_name_path = pathlib.Path( f"camp_{counter}" ) # pathlib.Path( out_name+f"_{counter}" )
        counter += 1
        #dot.render( directory=str(out_name_path.parent), filename=str(out_name_path.name) )
        subgraph = dot.subgraph(cluster)
        cluster_dot = gv.Digraph(out_name, comment='Patient Pathway Design', format="svg")
        for connection in connections:
            from_event = connection["from"]
            to_event = connection["to"]
            if event_names[from_event] in cluster:
                print( "Recreating edge in subgraph." )
                #node_color = "white" #get_colour_from_event(pieces)
                #node_shape = "ellipse" #get_shape_from_event(pieces)

                #subgraph.add_node( name=event_names[from_event], style='filled', fillcolor = node_color, shape=node_shape )
                #subgraph.add_edge( str(event_names[from_event]), str(event_names[to_event]), label=event )
                #cluster_dot.node( name=event_names[event_num], style='filled', fillcolor = node_color, shape=node_shape )
                cluster_dot.edge( str(event_names[from_event]), str(event_names[to_event]), label=connection["signal"] )
        cluster_dot.render( directory=str(out_name_path.parent), filename=str(out_name_path.name) )
        """
        #pos = graphviz.nx_agraph.graphviz_layout(subgraph, prog='dot')
        pos = graphviz.shell_layout(subgraph)
        graphviz.draw( subgraph, pos=pos, with_labels=True )
        plt.show()
        #plt.savefig( out_name_path )
        """
 
    return out_name
    

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--ccdl', help='Path to existing campaign in CCDL format.', default= "campaign.ccdl" )
    parser.add_argument('-w', '--whitelist', help='Optional trigger to limit visualization to.', default=None )
    parser.add_argument('-o', '--output', help='Path of output sv file.', default="camp.sv" )
    parser.add_argument('-d', '--display', help='Try to display', default=True )

    args = parser.parse_args()
    #  in_name = "campaign.ccdl", out_name = "camp.sv", display = True, whitelist = None ): 
    viz( in_name=args.ccdl, out_name=args.output, display=args.display, whitelist=args.whitelist )
